#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#ifdef HAVE_IMAGE_MAGICK
# include <MagickWand/MagickWand.h>
#endif
#ifdef IMGWORKS_USE_THREADS
# include <pthread.h>
#endif
#include <signal.h>
#include <gtk/gtk.h>
#include <gdk/gdkrgb.h>
#if defined(_WIN32)
# include <gdk/gdkwin32.h>
# define HAVE_WIN32
#else
# include <gdk/gdkx.h>
# define HAVE_X
#endif
#ifdef HAVE_LIBENDEAVOUR2
#include <endeavour2.h>
#endif

#include "guiutils.h"
#include "fprompt.h"
#include "cdialog.h"
#include "fb.h"
#include "pdialog.h" 
#include "csd.h"

#include "cfg.h"
#include "cfg_fio.h"
#include "obj.h"
#include "win.h"
#include "winlist.h"
#include "core.h"
#include "cfglist.h"
#include "config.h"


static void MainSignalCB(int s);
static void MainFileSelectorCreatedCB(
	const gchar *path, gpointer data
);
static void MainFileSelectorModifiedCB(
	const gchar *old_path,
	const gchar *new_path,
	gpointer data
);
static void MainFileSelectorDeletedCB(
	const gchar *path, gpointer data
);
static gint MainTOCB(gpointer data);

static CfgItem *MainCfgListNew(void);

win_struct *MainNewImgWorks(core_struct *core, gint argc, gchar **argv); 


#ifdef HAVE_IMAGE_MAGICK
#define MAGICK_FREE(p)		(((p) != NULL) ? \
 MagickRelinquishMemory(p) : NULL)
#endif

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	UNIX signal callback.
 */
static void MainSignalCB(int s)
{
	switch(s)
	{
	  case SIGINT:
	  case SIGTERM:
	  case SIGABRT:
		exit(4);	/* User aborted */
		break;

	  case SIGSEGV:
		exit(8);	/* Segmentation fault */
		break;

	  case SIGSTOP:
		break;

	  case SIGCONT:
		break;
	}
}

/*
 *	File selector created callback.
 */
static void MainFileSelectorCreatedCB(
	const gchar *path, gpointer data
)
{
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *ctx = EDV_CONTEXT(data);
	edv_notify_queue_vfs_object_added(ctx, path);
	edv_context_flush(ctx);
#endif
}

/*
 *	File selector changed callback.
 */
static void MainFileSelectorModifiedCB(
	const gchar *old_path,
	const gchar *new_path,
	gpointer data
)
{
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *ctx = EDV_CONTEXT(data);
	edv_notify_queue_vfs_object_modified(ctx, old_path, new_path);
	edv_context_flush(ctx);
#endif
}

/*
 *	File selector deleted callback.
 */
static void MainFileSelectorDeletedCB(
	const gchar *path, gpointer data
)
{
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *ctx = EDV_CONTEXT(data);
	edv_notify_queue_vfs_object_removed(ctx, path);
	edv_context_flush(ctx);
#endif
}


/*
 *	Main GTK timeout callback.
 */
static gint MainTOCB(gpointer data)
{
	gint total_windows = 0;
	GList *glist;
	win_struct *win;
	core_struct *core = CORE(data);
	if(core == NULL)
		return(FALSE);

	gdk_threads_enter();

	/* Manage all windows, count the number of windows that are
	 * mapped and delete any unmapped windows
	 */
	for(glist = core->win;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		win = WIN(glist->data);
		if(win == NULL)
			continue;

		if(core->close_all_windows && win->processing)
			core->close_all_windows = FALSE;

		if(WinIsMapped(win))
		{
			total_windows++;
		}
		else
		{
			WinDelete(win);
			glist->data = NULL;
			continue;
		}
	}

	/* If no windows are mapped then break out of the main loop */
	if((total_windows == 0) || core->close_all_windows)
	{
		gtk_main_quit();
		gdk_threads_leave();
		return(FALSE);
	}
	gdk_threads_leave();

	return(TRUE);
}


/*
 *	Creates a new default Configuration List.
 */
static CfgItem *MainCfgListNew(void)
{
	const CfgItem src_cfg_list[] = CFG_LIST;
	CfgItem *cfg_list = CFGItemListCopyList(src_cfg_list);
	gchar *cwd = STRDUP(g_get_current_dir());

	/* Version */
	CFG_SET_I(CFG_PARM_VERSION_MAJOR, PROG_VERSION_MAJOR);
	CFG_SET_I(CFG_PARM_VERSION_MINOR, PROG_VERSION_MINOR);
	CFG_SET_I(CFG_PARM_VERSION_RELEASE, PROG_VERSION_RELEASE);

	/* Convert Options */
	CFG_SET_S(CFG_PARM_TAR_DIR, cwd);
	CFG_SET_B(CFG_PARM_REMOVE_CONVERTED, TRUE);
	CFG_SET_B(CFG_PARM_RECURSIVE, FALSE);

	/* Win Options */
	CFG_SET_I(CFG_PARM_WIN_X, 0);
	CFG_SET_I(CFG_PARM_WIN_Y, 0);
	CFG_SET_I(CFG_PARM_WIN_WIDTH, -1);
	CFG_SET_I(CFG_PARM_WIN_HEIGHT, -1);
	CFG_SET_I(CFG_PARM_WIN_PANED1_POSITION, -1);
	CFG_SET_I(CFG_PARM_WIN_PANED2_POSITION, -1);
	CFG_SET_B(CFG_PARM_WIN_PREVIEW_SRC, TRUE);
	CFG_SET_B(CFG_PARM_WIN_PREVIEW_SRC_SHOW_DETAILS, FALSE);
	CFG_SET_B(CFG_PARM_WIN_PREVIEW_TAR, TRUE);
	CFG_SET_B(CFG_PARM_WIN_PREVIEW_TAR_SHOW_DETAILS, FALSE);


	g_free(cwd);

	return(cfg_list);
}


/*
 *	Creates a new ImgWorks and adds it to the list.
 */
win_struct *MainNewImgWorks(core_struct *core, gint argc, gchar **argv)
{
	GList *glist;
	win_struct *win = WinNew(core);
	if(win == NULL)
		return(NULL);

	GUIWindowApplyArgs(GTK_WINDOW(win->toplevel), argc, argv);

	WinMap(win);

	/* Add pointer to the list */
	for(glist = core->win;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		if(glist->data == NULL)
		{
			glist->data = win;
			return(win);
		}
	}
	core->win = g_list_append(core->win, win);
	return(win);
}


int main(int argc, char *argv[])
{
	gboolean initialized_gtk = FALSE;
	gint i;
	const gchar *arg;
	gchar *cfg_file = NULL;
	GList *paths_list = NULL;
#ifdef HAVE_X
	Display *dpy;
#endif
	GdkWindow *root;
	CfgItem *cfg_list;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *ctx;
#endif
	win_struct *win;
	core_struct *core;


	/* Set up time zone */
	tzset();

	/* Set up signal callbacks */
	signal(SIGINT, MainSignalCB);
	signal(SIGTERM, MainSignalCB);
	signal(SIGABRT, MainSignalCB);
	signal(SIGSEGV, MainSignalCB);
	signal(SIGSTOP, MainSignalCB);
	signal(SIGCONT, MainSignalCB);
	signal(SIGPIPE, MainSignalCB);


	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
		arg = argv[i];
		if(STRISEMPTY(arg))
			continue;

		/* Help */
		if(!g_strcasecmp(arg, "--help") ||
		   !g_strcasecmp(arg, "-help") ||
		   !g_strcasecmp(arg, "--h") ||
		   !g_strcasecmp(arg, "-h") ||
		   !g_strcasecmp(arg, "-?")
		)
		{
			g_print("%s", PROG_USAGE_MESG);
			g_list_foreach(paths_list, (GFunc)g_free, NULL);
			g_list_free(paths_list);
			return(0);
		}
		/* Version */
		else if(!g_strcasecmp(arg, "--version") ||
				!g_strcasecmp(arg, "-version") ||
				!g_strcasecmp(arg, "--v") ||
				!g_strcasecmp(arg, "-v")
		)
		{
			g_print(
				PROG_NAME_FULL " Version " PROG_VERSION "\n"
				PROG_COPYRIGHT
			);
			g_print("\nSupported Libraries:\n");
#ifdef HAVE_IMAGE_MAGICK
			if(TRUE)
			{
				unsigned long n;
				const gchar *s = (const gchar *)MagickGetVersion(&n);
				g_print("\t%s\n", s);
			}
#endif
			g_list_foreach(paths_list, (GFunc)g_free, NULL);
			g_list_free(paths_list);
			return(0);
		}
		/* Formats */
		else if(!g_strcasecmp(arg, "--formats") ||
				!g_strcasecmp(arg, "-formats")
		)
		{
#ifdef HAVE_IMAGE_MAGICK
			unsigned long n;
			char **strv;
			gint i;
			strv = MagickQueryFormats("*", &n);
			for(i = 0; i < (gint)n; i++)
			{
				g_print("%s\n", (const gchar *)strv[i]);
				MAGICK_FREE(strv[i]);
			}
			MAGICK_FREE(strv);
#endif
			g_list_foreach(paths_list, (GFunc)g_free, NULL);
			g_list_free(paths_list);
			return(0);
		}
		/* Skip GUI arguments that have a an argument */
		else if(!g_strcasecmp(arg, "--geometry") ||
				!g_strcasecmp(arg, "-geometry") ||
				!g_strcasecmp(arg, "--fn") ||
				!g_strcasecmp(arg, "-fn") ||
				!g_strcasecmp(arg, "--fg") ||
				!g_strcasecmp(arg, "-fg") ||
				!g_strcasecmp(arg, "--bg") ||
				!g_strcasecmp(arg, "-bg") ||
				!g_strcasecmp(arg, "--sfg") ||
				!g_strcasecmp(arg, "-sfg") ||
				!g_strcasecmp(arg, "--sbg") ||
				!g_strcasecmp(arg, "-sbg") ||
				!g_strcasecmp(arg, "--cfg") ||
				!g_strcasecmp(arg, "-cfg") ||
				!g_strcasecmp(arg, "--cbg") ||
				!g_strcasecmp(arg, "-cbg") ||
				!g_strcasecmp(arg, "--bg-pixmap") ||
				!g_strcasecmp(arg, "-bg-pixmap") ||
				!g_strcasecmp(arg, "--sbg-pixmap") ||
				!g_strcasecmp(arg, "-sbg-pixmap") ||
				!g_strcasecmp(arg, "--name") ||
				!g_strcasecmp(arg, "-name") ||
				!g_strcasecmp(arg, "--class") ||
				!g_strcasecmp(arg, "-class") ||
				!g_strcasecmp(arg, "--title") ||
				!g_strcasecmp(arg, "-title")
		)
		{
			i++;
		}
		/* Path */
		else if((*arg != '-') && (*arg != '+'))
		{
			paths_list = g_list_append(paths_list, STRDUP(arg));
		}
	}


#ifdef IMGWORKS_USE_THREADS
	/* Initialize GLIB threads mutex */
	g_thread_init(NULL);
#endif

	/* Initialize GTK as needed */
	if(!initialized_gtk)
	{
		if(!gtk_init_check(&argc, &argv))
		{
			g_printerr(
"Unable to initialize GTK.\n"
			);
			g_list_foreach(paths_list, (GFunc)g_free, NULL);
			g_list_free(paths_list);
			return(1);
		}

		/* Initialize the GDK RGB buffers system */
		gdk_rgb_init();

		initialized_gtk = TRUE;
	}

#ifdef HAVE_X
	dpy = GDK_DISPLAY();
#endif
	root = GDK_ROOT_PARENT();

	/* Initialize the dialogs */
	FPromptInit();
	CDialogInit();
	FileBrowserInit();
	PDialogInit();
	CSDInit();

	/* Create the core */
	core = CORE(g_malloc0(sizeof(core_struct)));
	if(core == NULL)
	{
		g_list_foreach(paths_list, (GFunc)g_free, NULL);
		g_list_free(paths_list);
		return(3);
	}

	/* Configuration List */
	core->cfg_list = cfg_list = MainCfgListNew();

	/* Cursors */
	core->busy_cur = gdk_cursor_new(GDK_WATCH);
	if(TRUE)
	{
		GdkColor    c_fg = { 0l, 0xffff, 0xffff, 0xffff },
					c_bg = { 0l, 0x0000, 0x0000, 0x0000 };
		const gint width = 24, height = 24;
		const gchar color_mask_data[] = {
0x03, 0x00, 0x00,
0x05, 0x00, 0x00,
0x09, 0x00, 0x00,
0x11, 0x00, 0x00,
0x21, 0x00, 0x00,
0x41, 0x00, 0x00,
0x81, 0x00, 0x00,
0x01, 0x01, 0x00,
0x01, 0xc2, 0x7f,
0xc1, 0x03, 0x00,
0x49, 0x80, 0x3f,
0x95, 0x80, 0x3f,
0x93, 0x80, 0x35,
0x20, 0x01, 0x1b,
0x20, 0x01, 0x0e,
0xc0, 0x00, 0x04,
0x00, 0x00, 0x0a,
0x00, 0x00, 0x1f,
0x00, 0x80, 0x3b,
0x00, 0x80, 0x35,
0x00, 0x80, 0x2a,
0x00, 0x00, 0x00,
0x00, 0xc0, 0x7f,
0x00, 0x00, 0x00
		};
		const gchar trans_mask_data[] = {
0x03, 0x00, 0x00,
0x07, 0x00, 0x00,
0x0f, 0x00, 0x00,
0x1f, 0x00, 0x00,
0x3f, 0x00, 0x00,
0x7f, 0x00, 0x00,
0xff, 0x00, 0x00,
0xff, 0xe1, 0xff,
0xff, 0xe3, 0xff,
0xff, 0xe3, 0xff,
0x7f, 0xc0, 0x7f,
0xf7, 0xc0, 0x7f,
0xf3, 0xc0, 0x7f,
0xe0, 0x81, 0x3f,
0xe0, 0x01, 0x1f,
0xc0, 0x00, 0x0e,
0x00, 0x00, 0x1f,
0x00, 0x80, 0x3f,
0x00, 0xc0, 0x7f,
0x00, 0xc0, 0x7f,
0x00, 0xc0, 0x7f,
0x00, 0xe0, 0xff,
0x00, 0xe0, 0xff,
0x00, 0xe0, 0xff
		};
		GdkBitmap *color_mask = gdk_bitmap_create_from_data(
			root, color_mask_data, width, height
		);
		GdkBitmap *trans_mask = gdk_bitmap_create_from_data(
			root, trans_mask_data, width, height
		);
		core->passive_busy_cur = gdk_cursor_new_from_pixmap(
			color_mask, trans_mask,
			&c_fg, &c_bg,
			0, 0
		);
		GDK_BITMAP_UNREF(color_mask);
		GDK_BITMAP_UNREF(trans_mask);
	}
	else
	{
		core->passive_busy_cur = gdk_cursor_new(GDK_WATCH);
	}
	core->translate_cur = gdk_cursor_new(GDK_FLEUR);

	core->close_all_windows = FALSE;

	core->win = NULL;

#ifdef HAVE_LIBENDEAVOUR2
	/* Create the Endeavour Mark II context */
	core->edv_ctx = ctx = edv_context_new();
	edv_context_init(ctx, NULL);
#endif

	/* Set the file selector callbacks */
	FileBrowserSetObjectCreatedCB(
		MainFileSelectorCreatedCB,
#ifdef HAVE_LIBENDEAVOUR2
		ctx
#else
		NULL
#endif
	);
	FileBrowserSetObjectModifiedCB(
		MainFileSelectorModifiedCB,
#ifdef HAVE_LIBENDEAVOUR2
		ctx
#else
		NULL
#endif
	);
	FileBrowserSetObjectDeletedCB(
		MainFileSelectorDeletedCB,
#ifdef HAVE_LIBENDEAVOUR2
		ctx
#else
		NULL
#endif
	);

	/* Get the configuration file location */
#ifdef HAVE_LIBENDEAVOUR2
	if(cfg_file == NULL)
		cfg_file = g_strconcat(
			edv_get_s(ctx, EDV_CFG_PARM_DIR_LOCAL),
			"/",
			"imgworks.ini",
			NULL
		);
#endif
	if(cfg_file == NULL)
		cfg_file = g_strconcat(
			g_getenv("HOME"),
			"/",
			".imgworks",
			NULL
		);

	/* Open the configuration from file */
	CFGFileOpen(cfg_file, cfg_list);


	/* Create the first ImgWorks window */
	win = MainNewImgWorks(core, argc, argv);
	if(paths_list != NULL)
	{
		const gchar *path;
		gchar *cwd = STRDUP(g_get_current_dir());
		GList *glist;
		obj_struct *obj;

		for(glist = paths_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
			path = (const gchar *)glist->data;
			if(path == NULL)
				continue;

			obj = ObjNew();
			if(obj != NULL)
			{
				if(g_path_is_absolute(path))
				{
					obj->name = STRDUP(g_basename(path));
					obj->full_path = STRDUP(path);
				}
				else
				{
					obj->full_path = g_strdup_printf(
						"%s%c%s",
						cwd, G_DIR_SEPARATOR, path
					);
					obj->name = STRDUP(g_basename(obj->full_path));
				}
				WinSrcListAppend(win, obj);
			}
		}
		WinUpdate(win);

		g_free(cwd);

		g_list_foreach(paths_list, (GFunc)g_free, NULL);
		g_list_free(paths_list);
		paths_list = NULL;
	}

	/* Set the main GTK timeout callback */
	gtk_timeout_add(
		1000l,				/* Delay in milliseconds */
		MainTOCB, core
	);
	gdk_threads_enter();
	gtk_main();
	gdk_threads_leave();

	/* Save the configuration to file */
	CFGFileSave(cfg_file, cfg_list);


	/* Delete the core */
	g_list_foreach(core->win, (GFunc)WinDelete, NULL);
	g_list_free(core->win);
	core->win = NULL;
#ifdef HAVE_LIBENDEAVOUR2
	/* Endeavour Mark II context */
	edv_context_delete(core->edv_ctx);
#endif
	GDK_CURSOR_DESTROY(core->busy_cur);
	GDK_CURSOR_DESTROY(core->passive_busy_cur);
	GDK_CURSOR_DESTROY(core->translate_cur);
	CFGItemListDeleteList(core->cfg_list);
	g_free(core);

	g_free(cfg_file);

	/* Shutdown dialogs */
	CSDInit();
	PDialogShutdown();
	FileBrowserShutdown();
	CDialogShutdown();
	FPromptShutdown();

	return(0);
}
