#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef IMGWORKS_USE_THREADS
# include <pthread.h>
#endif
#if defined(HAVE_IMAGE_MAGICK)
# include <MagickWand/MagickWand.h>
#endif
#include <gtk/gtk.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif
#include <unistd.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "cdialog.h"
#include "animicon.h"

#include "obj.h"
#include "convparm.h"
#include "win.h"
#include "winlist.h"
#include "winpreview.h"
#include "winconv.h"
#include "core.h"

#include "images/icon_replace_file_32x32.xpm"


/*
 *	Conversion Context:
 *
 *	Holds conversion library-specific resources.
 */
typedef struct {
#if defined(HAVE_IMAGE_MAGICK)
#endif
	gchar		*tar_fmt;

} conv_context_struct;


/* Utils */
gboolean WinConvertIsImageSupported(
	win_struct *win, const gchar *path  
);
static gchar *WinConvertFormatTargetPath(
	conv_context_struct *ctx,
	obj_struct *obj, const gchar *tar_dir
);
static gint WinConvertQueryOverwrite(
	core_struct *core, win_struct *win,
	const gchar *path, const struct stat *lstat_buf,
	GtkWidget *toplevel
);
static void WinConvertUpdatePreview(
	win_struct *win,
	const gint prev_sel_src,
	const gint prev_sel_tar
);

/* Callbacks */
#if defined(HAVE_IMAGE_MAGICK)
static MagickBooleanType WinConvertImageMagickProgressCB(
	const char *text,
	const MagickOffsetType offset,
	const MagickSizeType span,
	void *data
);
#endif

/* Conversion Context */
static conv_context_struct *WinConvertConvContextNew(void);
static void WinConvertConvContextDelete(conv_context_struct *ctx);

/* Convert Iteration */
#if defined(HAVE_IMAGE_MAGICK)
static gint WinConvertMagickIterate(
	conv_context_struct *ctx,
	core_struct *core, win_struct *win,
	const gchar *src_path,		/* Source object to convert */
	const gchar *tar_path,		/* Target object to write to */
	const struct stat *src_stat,
	GtkWidget *toplevel
);
#endif
static gint WinConvertIterate(
	conv_context_struct *ctx,
	core_struct *core, win_struct *win, obj_struct *obj,
	const gchar *tar_dir,
	gboolean recursive, gboolean *yes_to_all, gint *nconverted,
	GtkWidget *toplevel
);

/* Conversion Start/End Setup */
static void WinConvertSetStart(win_struct *win);
static void WinConvertSetEnd(win_struct *win);

/* Conversion Parameters to Conversion Context */
static void WinConvertSetParametersToContext(
	conv_context_struct *ctx, GList *convparm_list
);

void WinConvert(win_struct *win);
void WinConvertSelected(win_struct *win);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

#ifdef HAVE_IMAGE_MAGICK
#define MAGICK_FREE(p)		(((p) != NULL) ? \
 MagickRelinquishMemory(p) : NULL)
#define DESTROY_MAGICK_WAND(p)	(((p) != NULL) ? \
 DestroyMagickWand(p) : NULL)
#define DESTROY_PIXEL_WAND(p)	(((p) != NULL) ? \
 DestroyPixelWand(p) : NULL)

#define GDK_COLOR_TO_PIXEL_WAND(_c_,_wand_)	\
{ if(((_c_) != NULL) && ((_wand_) != NULL)) {	\
 PixelSetRed(					\
  (_wand_),					\
  (double)((gfloat)(_c_)->red / (gfloat)((gushort)-1)) \
 );						\
 PixelSetGreen(					\
  (_wand_),					\
  (double)((gfloat)(_c_)->green / (gfloat)((gushort)-1)) \
 );						\
 PixelSetBlue(					\
  (_wand_),					\
  (double)((gfloat)(_c_)->blue / (gfloat)((gushort)-1)) \
 );						\
 PixelSetAlpha(					\
  (_wand_),					\
  1.0						\
 );						\
} }

#endif	/* HAVE_IMAGE_MAGICK */

#if defined(HAVE_IMAGE_MAGICK) && defined(HAVE_LIBENDEAVOUR2)
#define IMAGE_MAGICK_THROW_EXCEPTION(_wand_)	{	\
 ExceptionType severity;				\
 char *description = MagickGetException(		\
  (_wand_), &severity					\
 );							\
 gchar *s = g_strdup_printf(				\
  "%s %s %ld %s\n",					\
  GetMagickModule(),					\
  description						\
 );							\
 edv_play_sound_warning(edv_ctx);				\
 CDialogSetTransientFor(toplevel);			\
 CDialogGetResponse(					\
  "Image Magick Exception",				\
  s,							\
  NULL,							\
  CDIALOG_ICON_WARNING,					\
  CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK		\
 );							\
 CDialogSetTransientFor(NULL);				\
 g_free(s);						\
 description = (char *)MAGICK_FREE(			\
  description						\
 );							\
}
#elif defined(HAVE_IMAGE_MAGICK)
#define IMAGE_MAGICK_THROW_EXCEPTION(_wand_)	{	\
 ExceptionType severity;				\
 char *description = MagickGetException(                \
  (_wand_), &severity                                   \
 );							\
 gchar *s = g_strdup_printf(				\
  "%s %s %ld %s\n",					\
  GetMagickModule(),					\
  description						\
 );							\
 CDialogSetTransientFor(toplevel);			\
 CDialogGetResponse(					\
  "Image Magick Exception",				\
  s,							\
  NULL,							\
  CDIALOG_ICON_WARNING,					\
  CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK		\
 );							\
 CDialogSetTransientFor(NULL);				\
 g_free(s);						\
 description = (char *)MAGICK_FREE(			\
  description						\
 );							\
}
#endif


/*
 *	Checks if the specified image is supported.
 */
gboolean WinConvertIsImageSupported(
	win_struct *win, const gchar *path  
)
{
#if defined(HAVE_IMAGE_MAGICK)
	unsigned long n;
	char **strv;
	gboolean status = FALSE;
	gint i;
	const gchar *ext;

	if(STRISEMPTY(path))
		return(status);

	ext = strrchr(path, '.');
	if(ext == NULL)
		return(status);

	ext++;

	strv = MagickQueryFormats("*", &n);
	for(i = 0; i < (gint)n; i++)
	{
		if(!status)
		{
			if(!g_strcasecmp(ext, strv[i]))
				status = TRUE;
		}

		MAGICK_FREE(strv[i]);
	}

	MAGICK_FREE(strv);

	return(status);
#else
	return(FALSE);
#endif
}


/*
 *	Formats the target path based on the specified object, format,
 *	and target directory.
 *
 *	The calling function must delete the returned string.
 */
static gchar *WinConvertFormatTargetPath(
	conv_context_struct *ctx,
	obj_struct *obj, const gchar *tar_dir
)
{
	gchar	*rtn_path, *ext = NULL,
			*name_no_ext = NULL, *name_postfix = NULL, *parent;

	if(STRISEMPTY(obj->full_path))
		return(NULL);

	/* Get the target directory and name postfix (if needed) */
	if(STRISEMPTY(tar_dir))
	{
		if(ConvParmFmtCompare(obj->full_path, ctx->tar_fmt))
			name_postfix = STRDUP("_conv");
		parent = g_dirname(obj->full_path);
	}
	else
	{
		parent = STRDUP(tar_dir);
	}
	if(STRISEMPTY(parent))
	{
		g_free(name_postfix);
		g_free(parent);
		return(NULL);
	}

	/* Get the name without the extension */
	if(obj->name != NULL)
	{
		gchar *s;
		name_no_ext = STRDUP(obj->name);
		s = strrchr(name_no_ext, '.');
		if(s != NULL)
			*s = '\0';
	}

	/* Get the target extension */
	ext = STRDUP(ctx->tar_fmt);
	strtolower(ext);
	if((ext != NULL) ? !g_strcasecmp(ext, "*") : FALSE)
	{
		const gchar *s = strrchr(obj->full_path, '.');
		g_free(ext);
		ext = (s != NULL) ? STRDUP(s + 1) : NULL;
		strtolower(ext);
	}

	/* Format the target path */
	rtn_path = g_strdup_printf(
		"%s%c%s%s.%s",
		parent,
		G_DIR_SEPARATOR,
		(name_no_ext != NULL) ? name_no_ext : "unnamed",
		(name_postfix != NULL) ? name_postfix : "",
		(ext != NULL) ? ext : ""
	);

	g_free(ext);
	g_free(name_no_ext);
	g_free(name_postfix);
	g_free(parent);

	return(rtn_path);
}

/*
 *	Queries overwrite of the specified object.
 */
static gint WinConvertQueryOverwrite(
	core_struct *core, win_struct *win,
	const gchar *path, const struct stat *lstat_buf,
	GtkWidget *toplevel
)
{
	gint response;
	const gulong	size = (gulong)lstat_buf->st_size,
					mtime = (gulong)lstat_buf->st_mtime;
	gchar *s;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx = core->edv_ctx;
	gchar	*size_str = STRDUP(edv_str_size_delim(size)),
			*date_str = STRDUP(edv_date_format(edv_ctx, mtime));
#else
	time_t t = mtime;
	gchar	*size_str = g_strdup_printf("%ld", size),
			*date_str = STRDUP(ctime(&t));
	if(date_str != NULL)
	{
		s = strchr(date_str, '\n');
		if(s != NULL)
			*s = '\0';
	}
#endif
	s = g_strdup_printf(
"Overwrite existing file:\n\
\n\
    %s  (%s %s)  %s",
		path,
		size_str,
		(size == 1l) ? "byte" : "bytes",
		date_str
	);
	g_free(size_str);
	g_free(date_str);
#ifdef HAVE_LIBENDEAVOUR2
	edv_play_sound_warning(edv_ctx);
#endif
	CDialogSetTransientFor(toplevel); 
	response = CDialogGetResponseIconData(
		"Confirm Overwrite",
		s,
		NULL,
		(guint8 **)icon_replace_file_32x32_xpm,
		CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_YES_TO_ALL |
		CDIALOG_BTNFLAG_NO | CDIALOG_BTNFLAG_CANCEL,
		CDIALOG_BTNFLAG_NO
	);
	g_free(s);
	CDialogSetTransientFor(NULL);

	return(response);
}

/*
 *	Updates the displayed target preview as needed.
 *
 *	Checks if the currently selected source or target preview
 *	differs from the specified previous selected source or target
 *	preview and updates the displayed preview as needed.
 *
 *	Passing -1 for either prev_sel_src or prev_sel_tar will
 *	update the preview if the source or target lists has a new
 *	selected item.
 *
 *	Passing -2 for either prev_sel_src or prev_sel_tar will not
 *	update the preview for the source or target.
 */
static void WinConvertUpdatePreview(
	win_struct *win,
	const gint prev_sel_src,
	const gint prev_sel_tar
)
{
	GtkCList *clist;
	core_struct *core;

	if(win == NULL)
		return;

	core = CORE(win->core);
	if(core == NULL)
		return;

	clist = (GtkCList *)win->src_clist;
	if(win->preview_src &&
	   (g_list_length(clist->selection) == 1) &&
	   (prev_sel_src != -2)
	)
	{
		GList *glist = clist->selection;
		const gint sel_row = (glist != NULL) ? (gint)glist->data : -1;
		if(sel_row != prev_sel_src)
		{
			obj_struct *obj = OBJ(gtk_clist_get_row_data(clist, sel_row));
			if((obj != NULL) ? !STRISEMPTY(obj->full_path) : FALSE)
			{
				const char *path = obj->full_path;
				win->freeze_count++;
#ifdef HAVE_LIBENDEAVOUR2
				WinPreviewSetImageFile(
					win->src_preview, path,
					core->edv_ctx
				);
#else
				WinPreviewSetImageFile(
					win->src_preview, path
				);
#endif
				win->freeze_count--;
			}
		}
	}

	clist = (GtkCList *)win->tar_clist;
	if(win->preview_tar &&
	   (g_list_length(clist->selection) == 1) &&
	   (prev_sel_tar != -2)
	)
	{
		GList *glist = clist->selection;
		const gint sel_row = (glist != NULL) ? (gint)glist->data : -1;
		if(sel_row != prev_sel_tar)
		{
			obj_struct *obj = OBJ(gtk_clist_get_row_data(clist, sel_row));
			if((obj != NULL) ? !STRISEMPTY(obj->full_path) : FALSE)
			{
				const char *path = obj->full_path;
				win->freeze_count++;
#ifdef HAVE_LIBENDEAVOUR2
				WinPreviewSetImageFile(
					win->tar_preview, path,
					core->edv_ctx
				);
#else
				WinPreviewSetImageFile(
					win->tar_preview, path
				);
#endif
				win->freeze_count--;
			}
		}
	}
}

#if defined(HAVE_IMAGE_MAGICK)
/*
 *	Image Magick progress callback.
 */
static MagickBooleanType WinConvertImageMagickProgressCB(
	const char *text,
	const MagickOffsetType offset,
	const MagickSizeType span,
	void *data
)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return(MagickFalse);

	WinStatusProgress(
		win,
		((gulong)span > 0l) ?
			((gfloat)(offset + 1) / (gfloat)span) : -1.0f,
		TRUE
	);

	return(MagickTrue);
}
#endif


/*
 *	Creates a new conversion context;
 */
static conv_context_struct *WinConvertConvContextNew(void)
{
	conv_context_struct *ctx = (conv_context_struct *)g_malloc0(
		sizeof(conv_context_struct)
	);
	if(ctx == NULL)
		return(NULL);

	ctx->tar_fmt = STRDUP("*");

	return(ctx);
}

/*
 *	Deletes the conversion context.
 */
static void WinConvertConvContextDelete(conv_context_struct *ctx)
{
	if(ctx == NULL)
		return;

	g_free(ctx->tar_fmt);
	g_free(ctx);
}



#if defined(HAVE_IMAGE_MAGICK)
/*
 *	Uses ImageMagick to convert the image specified by src_path.
 *
 *	All inputs assumed valid.
 *
 *	Return values:
 *
 *	0	Success.
 *	-1	Error.
 *	-2	Bad value.
 *	-3	Systems error.
 *	-4	User responded with "cancel".
 *	-5	User responded with "no".
 */
static gint WinConvertMagickIterate(
	conv_context_struct *ctx,
	core_struct *core, win_struct *win,
	const gchar *src_path,		/* Source object to convert */
	const gchar *tar_path,		/* Target object to write to */
	const struct stat *src_stat,
	GtkWidget *toplevel
)
{
	const gulong size = (gulong)src_stat->st_size;
	gint	status = 0,
			orig_width, orig_height, orig_nframes,
			width, height, frame_num, nframes;
	gchar *s, *format_name, *size_str;
	MagickWand *wand;
	PixelWand	*fg_pixel_wand,
					*bg_pixel_wand;
	GList *glist;
	convparm_struct *cp;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx = core->edv_ctx;
#endif

	/* User aborted? */
	if(win->stop_count > 0)
		return(-4);

	/* Create Image Magick context */
	wand = NewMagickWand();
	if(wand == NULL)
		return(-3);

	MagickSetProgressMonitor(
		wand,
		WinConvertImageMagickProgressCB,
		win
	);

	/* Print reading status message */
#ifdef HAVE_LIBENDEAVOUR2
	size_str = STRDUP(edv_str_size_delim(size));
#else
	size_str = g_strdup_printf("%ld", size);
#endif
	s = g_strdup_printf(
"Reading \"%s\" (%s %s)...",
		g_basename(src_path),
		size_str,
		(size == 1l) ? "byte" : "bytes"
	);
	g_free(size_str);
	WinStatusMessage(win, s, TRUE);
	g_free(s);

	/* Read image */
	if(MagickReadImage(wand, src_path) != MagickTrue)
	{
		/* Error reading image */
		IMAGE_MAGICK_THROW_EXCEPTION(wand);
		DESTROY_MAGICK_WAND(wand);
		return(-1);
	}

	/* User aborted? */
	if(win->stop_count > 0)
	{
		DESTROY_MAGICK_WAND(wand);
		return(-4);
	}

	/* Set up the foreground and background pixels */
	fg_pixel_wand = NULL;
	bg_pixel_wand = NewPixelWand();

	/* Get image information */
	orig_width = (gint)MagickGetImageWidth(wand);
	orig_height = (gint)MagickGetImageHeight(wand);
	orig_nframes = (gint)MagickGetNumberImages(wand);
	width = orig_width;
	height = orig_height;
	nframes = orig_nframes;

	/* Get the name of this image's format */
	format_name = MagickGetImageFormat(wand);

	/* Get the background color */
	if(bg_pixel_wand != NULL)
		MagickGetImageBackgroundColor(wand, bg_pixel_wand);

	/* Begin converting this image
	 *
	 * Iterate through each frame
	 */
	for(frame_num = 0; frame_num < nframes; frame_num++)
	{
		/* Begin converting this image */
		if(nframes > 1)
			s = g_strdup_printf(
"Converting \"%s\" %ix%i (frame %i/%i)...",
				g_basename(src_path),
				width, height,
				frame_num, nframes
			);
		else
			s = g_strdup_printf(
"Converting \"%s\" %ix%i...",
				g_basename(src_path),
				width, height
			);
		WinStatusMessage(win, s, TRUE);
		g_free(s);

		/* User aborted? */
		if(win->stop_count > 0)
			break;

		/* Iterate through each conversion parameter */
		for(glist = win->convparm_list;
			glist != NULL;
			glist = g_list_next(glist)
		)
		{
		    cp = CONVPARM(glist->data);
		    if(cp == NULL)
				continue;
		    if(!CONVPARM_IS_ENABLED(cp))
				continue;

			/* User aborted? */
			if(win->stop_count > 0)
				break;

		    /* Format */
		    if(CONVPARM_IS_FMT(cp))
		    {
				/* Already applied, skip */
		    }
		    /* Compression */
		    else if(CONVPARM_IS_COMPRESSION(cp))
		    {
				const gfloat compression = ConvParmCompressionGet(cp);
				if(compression <= 0.0f)
				{
				    if(MagickSetImageCompression(
						wand, NoCompression
				    ) == MagickFalse)
				    {
						IMAGE_MAGICK_THROW_EXCEPTION(wand);
						if(!status)
						    status = -1;
				    }
				}
				else if(!STRISEMPTY(ctx->tar_fmt))
				{
					const char *tar_fmt = ctx->tar_fmt;
					if(!g_strcasecmp(tar_fmt, "gif"))
					{
						if(MagickSetImageCompression(
							wand, RLECompression
						) == MagickTrue)
						{
							const unsigned long quality = (unsigned long)CLIP(
								compression * 100.0f,
								1.0f, 99.0f
							);
							MagickSetImageCompressionQuality(wand, quality);
						}
						else
						{
							IMAGE_MAGICK_THROW_EXCEPTION(wand);
							if(!status)
								status = -1;
						}
					}
					else if(!g_strcasecmp(tar_fmt, "jpeg") ||
						    !g_strcasecmp(tar_fmt, "jpg")
					)
					{
						if(MagickSetImageCompression(
							wand, JPEGCompression
						) == MagickTrue)
						{
							const unsigned long quality = (unsigned long)CLIP(
								100.0f - (compression * 100.0f),
								1.0f, 99.0f
							);
							MagickSetImageCompressionQuality(wand, quality);
						}
						else
						{
							IMAGE_MAGICK_THROW_EXCEPTION(wand);
							if(!status)
								status = -1;
						}
					}
					else if(!g_strcasecmp(tar_fmt, "png") ||
							!g_strcasecmp(tar_fmt, "png24") ||
							!g_strcasecmp(tar_fmt, "png32") ||
						    !g_strcasecmp(tar_fmt, "mng")
					)
					{
						if(MagickSetImageCompression(
							wand, 9
						) == MagickTrue)
						{
							const unsigned long quality = (unsigned long)CLIP(
								compression * 100.0f,
								1.0f, 99.0f
							);
							MagickSetImageCompressionQuality(wand, quality);
						}
						else
						{
							IMAGE_MAGICK_THROW_EXCEPTION(wand);
							if(!status)
								status = -1;
						}
					}
					else if(!g_strcasecmp(tar_fmt, "tga") ||
							!g_strcasecmp(tar_fmt, "tiff") ||
							!g_strcasecmp(tar_fmt, "tif")
					)
					{
						if(MagickSetImageCompression(
							wand, RLECompression
						) == MagickTrue)
						{
							const unsigned long quality = (unsigned long)CLIP(
								compression * 100.0f,
								1.0f, 99.0f
							);
							MagickSetImageCompressionQuality(wand, quality);
						}
						else
						{
							IMAGE_MAGICK_THROW_EXCEPTION(wand);
							if(!status)
								status = -1;
						}
					}
				}
		    }
		    /* Background Color */
		    else if(CONVPARM_IS_COLORBG(cp))
		    {
				PixelWand *pixel_wand = bg_pixel_wand;
				const GdkColor *c = ConvParmColorBGGet(cp);
				if((pixel_wand != NULL) && (c != NULL))
				{
				    GDK_COLOR_TO_PIXEL_WAND(c, pixel_wand);
				    if(MagickSetImageBackgroundColor(
						wand, pixel_wand
				    ) == MagickFalse)
				    {
						IMAGE_MAGICK_THROW_EXCEPTION(wand);
						if(!status)
						    status = -1;
				    }
				}
		    }
		    /* Gamma */
		    else if(CONVPARM_IS_GAMMA(cp))
		    {
				const gfloat g = ConvParmGammaGet(cp);
				if(MagickGammaImage(wand, (double)g)
				    == MagickFalse)
				{
				    IMAGE_MAGICK_THROW_EXCEPTION(wand);
				    if(!status)
						status = -1;
				}
		    }
		    /* Mirror */
		    else if(CONVPARM_IS_MIRROR(cp))
		    {
				const GtkOrientation orientation =
				    ConvParmMirrorGet(cp);
				switch(orientation)
				{
				  case GTK_ORIENTATION_VERTICAL:
				    if(MagickFlipImage(wand) == MagickFalse)
				    {
						IMAGE_MAGICK_THROW_EXCEPTION(wand);
						if(!status)
						    status = -1;
				    }
				    break;
				  case GTK_ORIENTATION_HORIZONTAL:
				    if(MagickFlopImage(wand) == MagickFalse)
				    {
						IMAGE_MAGICK_THROW_EXCEPTION(wand);
						if(!status)
						    status = -1;
				    }
				    break;
				}
				/* Get new image size after mirroring */
				width = (gint)MagickGetImageWidth(wand);
				height = (gint)MagickGetImageHeight(wand);
		    }
		    /* Crop */
		    else if(CONVPARM_IS_CROP(cp))
		    {
				const GdkRectangle *rect = ConvParmCropGet(cp);
				if(rect != NULL)
				{
				    if(MagickCropImage(
						wand,
						(unsigned long)rect->width,
						(unsigned long)rect->height,
						(long)rect->x,
						(long)rect->y
				    ) == MagickFalse)
				    {
						IMAGE_MAGICK_THROW_EXCEPTION(wand);
						if(!status)
						    status = -1;
				    }
				}
				/* Get new image size after cropping */
				width = (gint)MagickGetImageWidth(wand);
				height = (gint)MagickGetImageHeight(wand);
		    }
		    /* Rotate */
		    else if(CONVPARM_IS_ROTATE(cp))
		    {
				const gfloat deg = ConvParmRotateGet(cp);
				if(bg_pixel_wand != NULL)
				{
				    if(MagickRotateImage(
						wand,
						bg_pixel_wand,	/* Background */
						(double)deg		/* Angle in degrees */
				    ) == MagickFalse)
				    {
						IMAGE_MAGICK_THROW_EXCEPTION(wand);
						if(!status)
						    status = -1;
				    }
				    /* Get new image size after rotation since
				     * rotating may change the size
				     */
				    width = (gint)MagickGetImageWidth(wand);
				    height = (gint)MagickGetImageHeight(wand);
				}
		    }
		    /* Size */
		    else if(CONVPARM_IS_SIZE(cp))
		    {
				gint new_width, new_height;
				const convparm_sizeinfo_struct *s =
				ConvParmSizeGet(cp);
				if(s == NULL)
				    continue;
				if(s->maintain_aspect)
				{
				    const gfloat aspect = (height > 0) ?
						((gfloat)width / (gfloat)height) : 1.0;
				    if(s->smaller)
				    {
						new_height = s->height;
						new_width = new_height * aspect;
						if(new_width > s->width)
						{
						    new_width = s->width;
						    new_height = (aspect > 0.0f) ?
								(gint)(new_width / aspect) : 1;
						}
				    }
				    else
				    {
						new_height = s->height;
						new_width = new_height * aspect;
						if(new_width < s->width)
						{
						    new_width = s->width;
						    new_height = (aspect > 0.0f) ?
								(gint)(new_width / aspect) : 1;
						}
				    }
				}
				else
				{
				    new_width = s->width;
				    new_height = s->height;
				}
				if(MagickResizeImage(
				    wand,
				    (unsigned long)new_width,	/* New width */
				    (unsigned long)new_height,	/* New height */
				    LanczosFilter		/* Filter type */
				) == MagickFalse)
				{
				    IMAGE_MAGICK_THROW_EXCEPTION(wand);
				    if(!status)
						status = -1;
				}
				/* Get new image size after resizing */
				width = (gint)MagickGetImageWidth(wand);
				height = (gint)MagickGetImageHeight(wand);
		    }
		    /* Frame */
		    else if(CONVPARM_IS_FRAME(cp))
		    {
				PixelWand *pixel_wand = NewPixelWand();
				const convparm_frameinfo_struct *f =
				    ConvParmFrameGet(cp);
				if((f != NULL) && (pixel_wand != NULL))
				{
				    GDK_COLOR_TO_PIXEL_WAND(&f->c, pixel_wand);
				    if(MagickFrameImage(
						wand, pixel_wand,
						(unsigned long)f->width,
						(unsigned long)f->height,
						(long)f->inner_bevel,
						(long)f->outer_bevel,
						NoCompositeOp
				    ) == MagickFalse)
				    {
						IMAGE_MAGICK_THROW_EXCEPTION(wand);
						if(!status)
						    status = -1;
				    }
				}
				DESTROY_PIXEL_WAND(pixel_wand);
				/* Get new image size after adding a frame */   
				width = (gint)MagickGetImageWidth(wand);
				height = (gint)MagickGetImageHeight(wand);
		    }
		}	/* Iterate through each conversion parameter */

		MagickNextImage(wand);
	}

	/* User aborted? */
	if(win->stop_count > 0)
	{
		DESTROY_PIXEL_WAND(fg_pixel_wand);
		DESTROY_PIXEL_WAND(bg_pixel_wand);
		format_name = (gchar *)MAGICK_FREE(format_name);
		DESTROY_MAGICK_WAND(wand);
		return(-4);
	}

	/* Go back to the first frame */
	MagickSetIteratorIndex(wand, 0l);

	/* Print writing status message */
	s = g_strdup_printf(
"Writing \"%s\"...",
		g_basename(tar_path)
	);
	WinStatusMessage(win, s, TRUE);
	g_free(s);

	/* Write target image */
	if((nframes > 1) ?
		(MagickWriteImages(wand, tar_path, MagickTrue) != MagickFalse) :
		(MagickWriteImage(wand, tar_path) != MagickFalse)
	)
	{
		chmod(tar_path, src_stat->st_mode);
		lchown(tar_path, src_stat->st_uid, src_stat->st_gid);
	}
	else
	{
		/* Error writing image */
		IMAGE_MAGICK_THROW_EXCEPTION(wand);
		if(!status)
			status = -1;
	}

	DESTROY_PIXEL_WAND(fg_pixel_wand);
	DESTROY_PIXEL_WAND(bg_pixel_wand);
	format_name = (gchar *)MAGICK_FREE(format_name);

	DESTROY_MAGICK_WAND(wand);

	return(status);
}
#endif	/* HAVE_IMAGE_MAGICK */


/*
 *	Converts the image specified by the object.
 *
 *	If the specified object is a directory then each object in the
 *	directory will be converted. If the object within the directory
 *	is a subdirectory and recursive is TRUE then this function will
 *	be called again to recursively convert each object in that
 *	subdirectory.
 *
 *	Return values:
 *
 *	0	Success.
 *	-1	Error.
 *	-2	Bad value.
 *	-3	Systems error.
 *	-4	User responded with "cancel".
 *	-5	User responded with "no".
 */
static gint WinConvertIterate(
	conv_context_struct *ctx,
	core_struct *core, win_struct *win, obj_struct *obj,
	const gchar *tar_dir,
	gboolean recursive, gboolean *yes_to_all, gint *nconverted,
	GtkWidget *toplevel
)
{
	struct stat src_stat;
	gint status = 0;
	const gchar *src_path = obj->full_path;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx = core->edv_ctx;
#endif

	if(win->stop_count > 0)
		return(-4);

	if(STRISEMPTY(src_path))
		return(-2);

	if(stat((const char *)src_path, &src_stat))
		return(-1);

#ifdef S_ISDIR
	if(S_ISDIR(src_stat.st_mode))
#else
	if(FALSE)
#endif
	{
		gint nnames;
		gchar **names_list = (gchar **)GetDirEntNames2(
			(const char *)src_path, (int *)&nnames
		);
		if(names_list != NULL)
		{
			gint i;
			const gchar *name;
			gchar *path = NULL, *tar_subdir = NULL;

#define FREE_CONTINUE	{	\
 g_free(path);			\
 path = NULL;			\
 g_free(names_list[i]);		\
 names_list[i] = NULL;		\
 continue;			\
}

			/* If the target directory was specified then append
			 * the source directory's name to the new target
			 * subdirectory so that a new directory with the same
			 * name gets created at the target directory and the
			 * converted objects get put there
			 */
			if(!STRISEMPTY(tar_dir))
				tar_subdir = STRDUP(PrefixPaths(
					tar_dir,
					g_basename(src_path)
				));

			StringQSort((char **)names_list, (int)nnames);

			/* Iterate through each object in the source directory */
			for(i = 0; i < nnames; i++)
			{
				/* Canceled? */
				if(status == -4)
					FREE_CONTINUE

				/* User aborted? */
				if(win->stop_count > 0)
				{
					status = -4;
					FREE_CONTINUE
				}

				name = names_list[i];
				if(name == NULL)
					FREE_CONTINUE

				if(!g_strcasecmp(name, ".") ||
				   !g_strcasecmp(name, "..")
				)
					FREE_CONTINUE

				path = STRDUP(PrefixPaths(src_path, name));
				if(path == NULL)
					FREE_CONTINUE

				if(stat(path, &src_stat))
					FREE_CONTINUE

#ifdef S_ISDIR
				if(S_ISDIR(src_stat.st_mode))
#else
				if(FALSE)
#endif
				{
					/* This object leads to a directory, recurse
					 * into that directory?
					 */
					if(recursive)
					{
						obj_struct *obj = ObjNewFromPath(path);
						if(obj != NULL)
						{
							const gint status2 = WinConvertIterate(
								ctx, core, win, obj,
								tar_subdir,
								recursive, yes_to_all, nconverted,
								toplevel
							);
							if(status2)
								status = status2;
							ObjDelete(obj);
						}
					}
				}
				else if(WinConvertIsImageSupported(win, path))
				{
					obj_struct *obj = ObjNewFromPath(path);
					if(obj != NULL)
					{
						const gint status2 = WinConvertIterate(
						    ctx, core, win, obj,
						    tar_subdir,
						    recursive, yes_to_all, nconverted,
						    toplevel
						);
						if(status2)
						    status = status2;
						ObjDelete(obj);
					}
				}

				FREE_CONTINUE;
			}

			g_free(tar_subdir);
			g_free(names_list);
#undef FREE_CONTINUE
		}
	}
	else
	{
		/* Read, convert, and write this object */
		gboolean tar_existed = FALSE;
		gint prev_sel_src = -1, prev_sel_tar = -1;
		struct stat tar_lstat, tar_parent_stat;
		gchar *tar_path, *tar_parent;
		GList *glist;
		GtkCList *clist;

		/* Record the previously selected items */
		clist = (GtkCList *)win->src_clist;
		glist = clist->selection;
		if(g_list_length(glist) == 1)
			prev_sel_src = (gint)glist->data;
		clist = (GtkCList *)win->tar_clist;
		glist = clist->selection;
		if(g_list_length(glist) == 1)
			prev_sel_tar = (gint)glist->data;

		/* Get the path of the target object */
		tar_path = WinConvertFormatTargetPath(ctx, obj, tar_dir);
		if(tar_path == NULL)
			return(-1);

		/* Check if the target object's parent exist */
		tar_parent = g_dirname(tar_path);
		if((tar_parent != NULL) ?
			stat((char *)tar_path, &tar_parent_stat) : FALSE)
		{
#if defined(S_IRWXU) && defined(S_IRWXG) && defined(S_IRWXO)
			const mode_t m = S_IRWXU | S_IRWXG | S_IRWXO;
#else
			const mode_t m = 0;
#endif
			errno = 0;
			if(rmkdir((char *)tar_parent, m))
			{
				const gint error_code = (gint)errno;
				gchar *s = STRDUP(g_strerror(error_code));
				if(s != NULL)
					*s = (gchar)toupper((int)*s);
#ifdef HAVE_LIBENDEAVOUR2
				edv_play_sound_error(edv_ctx);
#endif
				CDialogSetTransientFor(toplevel);
				CDialogGetResponse(
					"Create Directory Failed",
					s,
					NULL,
					CDIALOG_ICON_ERROR,
					CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
				);
				g_free(s);
				CDialogSetTransientFor(NULL);
				g_free(tar_parent);
				g_free(tar_path);
				return(-1);
			}
			else
			{
				stat((const char *)tar_path, &tar_parent_stat);
#ifdef HAVE_LIBENDEAVOUR2
				edv_notify_queue_vfs_object_added(edv_ctx, tar_parent);
				edv_context_flush(edv_ctx);
#endif
			}
		}

		/* Check if the target object already exists */
		if(!lstat((const char *)tar_path, &tar_lstat))
		{
			/* Mark that the target object existed prior to the
			 * convert
			 */
			tar_existed = TRUE;

			/* Does it exist as a directory? */
#ifdef S_ISDIR
			if(S_ISDIR(tar_lstat.st_mode))
#else
			if(FALSE)
#endif
			{
#ifdef HAVE_LIBENDEAVOUR2
				edv_play_sound_warning(edv_ctx);
#endif
				CDialogSetTransientFor(toplevel);
				CDialogGetResponse(
					"Target Exists As A Directory",
"Unable to write target image, the target already\n\
exists and is a directory.",
					NULL,
					CDIALOG_ICON_WARNING,
					CDIALOG_BTNFLAG_OK,
					CDIALOG_BTNFLAG_OK
				);
				CDialogSetTransientFor(NULL);
				g_free(tar_parent);
				g_free(tar_path);
				return(-2);
			}
			/* Need to query the user to overwrite the target object? */
			else if(!(*yes_to_all))
			{
				/* Query the user to overwrite of the target object */
				const gint response = WinConvertQueryOverwrite(
					core, win, tar_path, &tar_lstat, toplevel
				);
				switch(response)
				{
				  case CDIALOG_RESPONSE_YES_TO_ALL:
					*yes_to_all = TRUE;
				  case CDIALOG_RESPONSE_YES:
					break;
				  case CDIALOG_RESPONSE_CANCEL:
					g_free(tar_parent);
					g_free(tar_path);
					return(-4);
					break;
				  default:	/* All else assume no */
					g_free(tar_parent);
					g_free(tar_path);
					return(-5);
					break;
				}
			}
		}	/* Check if the target path already exists */

#if defined(HAVE_IMAGE_MAGICK)
		/* Convert */
		status = WinConvertMagickIterate(
			ctx, core, win,
			src_path,		/* Source object to convert */
			tar_path,		/* Target object to write to */
			&src_stat,		/* Source stats */
			toplevel
		);
#endif

		/* Converted successfully? */
		if(!status)
		{
			/* Count this object as an image converted */
			*nconverted = (*nconverted) + 1;

			/* Append the converted image to the target list */
			WinTarListAppend(win, ObjNewFromPath(tar_path));

#ifdef HAVE_LIBENDEAVOUR2
			/* Notify about modified or added object */
			if(tar_existed)
				edv_notify_queue_vfs_object_modified(
					edv_ctx, tar_path, tar_path
				);
			else
				edv_notify_queue_vfs_object_added(
					edv_ctx, tar_path
				);
			edv_context_flush(edv_ctx);
#endif
		}

		/* Update the previews after converting this object
		 * as needed
		 */
		WinConvertUpdatePreview(win, prev_sel_src, prev_sel_tar);

		g_free(tar_parent);
		g_free(tar_path);
	}

	return(status);
}


/*
 *	Sets Win values for the start of conversion.
 */
static void WinConvertSetStart(win_struct *win)
{
	WinSetPassiveBusy(win, TRUE);
	win->processing = TRUE;
	win->stop_count = 0;
	AnimIconPlay(win->anim_icon);
	WinStatusProgress(win, 0.0f, TRUE);
	WinUpdate(win);
}

/*
 *	Sets Win values at the end of a conversion.
 */
static void WinConvertSetEnd(win_struct *win)
{
	win->processing = FALSE;
	win->stop_count = 0;
	AnimIconPause(win->anim_icon);
	AnimIconSeek(win->anim_icon, 0); 
	WinStatusProgress(win, 0.0f, TRUE);
	WinUpdate(win); 
	WinSetPassiveBusy(win, FALSE);
}


/*
 *	Sets up the conversion contents with the values currently set in
 *	the specified conversion parameters list.
 */
static void WinConvertSetParametersToContext(
	conv_context_struct *ctx, GList *convparm_list 
)
{
	GList *glist;
	convparm_struct *cp;

	/* Set values from each item in the conversion parameter list
	 * to the conversion context
	 */
	for(glist = convparm_list; glist != NULL; glist = g_list_next(glist))
	{
		cp = CONVPARM(glist->data);
		if(cp == NULL)
			continue;
		if(!CONVPARM_IS_ENABLED(cp))
			continue;

		/* Format */
		if(CONVPARM_IS_FMT(cp))
		{
			g_free(ctx->tar_fmt);
			ctx->tar_fmt = STRDUP(ConvParmFmtGet(cp));
		}
		/* Color Background */
		else if(CONVPARM_IS_COLORBG(cp))
		{

		}
	}
}


/*
 *	Convert.
 */
void WinConvert(win_struct *win)
{
	gboolean	remove_converted,
					recursive,
					yes_to_all = FALSE;
	gint i, status, nobjs, nconverted = 0;
	const gchar *tar_dir;
	GList *glist, *obj_list, *obj_converted_list;
	GtkWidget *toplevel;
	GtkCList *clist;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx;
#endif
	obj_struct *obj;
	core_struct *core;
	conv_context_struct *ctx;

	if(win == NULL)
		return;

	tar_dir = win->tar_dir;
	remove_converted = win->remove_converted;
	recursive = win->recursive;
	toplevel = win->toplevel;
	clist = (GtkCList *)win->src_clist;
	core = CORE(win->core);
	if((clist == NULL) || (core == NULL))
		return;

#ifdef HAVE_LIBENDEAVOUR2
	edv_ctx = core->edv_ctx;
#endif

	/* Get list of objects */
	obj_list = NULL;
	for(i = 0; i < clist->rows; i++)
		obj_list = g_list_append(
			obj_list,
			gtk_clist_get_row_data(clist, i)
		);
	if(obj_list == NULL)
	{
#ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_warning(edv_ctx);
#endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"No Images",
			"There are no images to convert.",
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		return;
	}

	/* Create a new conversion context */
	ctx = WinConvertConvContextNew();
	if(ctx == NULL)
	{
		g_list_free(obj_list);
		return;
	}

	/* Set the values of the conversion parameters to the
	 * conversion context
	 */
	WinConvertSetParametersToContext(ctx, win->convparm_list);

	/* Start converting */
	WinConvertSetStart(win);	

	WinStatusMessage(win, "Converting...", TRUE);

	/* Iterate through each object in the list */
	obj_converted_list = NULL;
	for(i = 0,
		nobjs = g_list_length(obj_list),
		glist = obj_list;
		glist != NULL;
		i++,
		glist = g_list_next(glist)
	)
	{
		if(win->stop_count > 0)
			break;

		obj = OBJ(glist->data);
		if(obj == NULL)
			continue;

		/* Convert this object */
		status = WinConvertIterate(
			ctx, core, win, obj,
			tar_dir,
			recursive, &yes_to_all, &nconverted,
			toplevel
		);
		if(status)
		{
			/* User aborted/canceled? */
			if(status == -4)
				break;
		}
		else
		{
			/* Converted successfully, add this object to the list  
			 * of successfully converted objects
			 */
			obj_converted_list = g_list_append(
				obj_converted_list, obj
			);
		}
	}

	/* Print conversion results */
	if(nconverted > 0)
	{
		gchar *s = g_strdup_printf(
			"Converted %i images%s",
			nconverted,
			win->stop_count ? " (interrupted)" : ""
		);
		WinStatusMessage(win, s, TRUE);
		g_free(s);
	}
	else
	{
		WinStatusMessage(
			win,
			"No images converted",
			TRUE
		);
	}

	/* Remove converted? */
	if((obj_converted_list != NULL) && remove_converted)
	{
		for(glist = obj_converted_list;
			glist != NULL;
			glist = g_list_next(glist)
		)
		{
			i = gtk_clist_find_row_from_data(clist, glist->data);
			if(i < 0)
				continue;

			WinSrcListRemove(win, i);
		}
	}

	/* End conversion */
	WinConvertConvContextDelete(ctx);
	g_list_free(obj_converted_list);
	g_list_free(obj_list);
	WinConvertSetEnd(win);

#ifdef HAVE_LIBENDEAVOUR2
	/* Play the convert completed sound */
	edv_play_sound_completed(edv_ctx);
#endif


}

/*
 *	Convert selected.
 */
void WinConvertSelected(win_struct *win)
{
	gboolean        remove_converted,
					recursive,
					yes_to_all = FALSE;
	gint i, status, nobjs, nconverted = 0;
	const gchar *tar_dir;
	GList *glist, *obj_list, *obj_converted_list;
	GtkWidget *toplevel;
	GtkCList *clist;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx;
#endif
	obj_struct *obj;
	core_struct *core;
	conv_context_struct *ctx;

	if(win == NULL)
		return;

	tar_dir = win->tar_dir;
	remove_converted = win->remove_converted;
	recursive = win->recursive;
	toplevel = win->toplevel;
	clist = (GtkCList *)win->src_clist;
	core = CORE(win->core);
	if((clist == NULL) || (core == NULL))
		return;

#ifdef HAVE_LIBENDEAVOUR2
	edv_ctx = core->edv_ctx;
#endif

	/* Get list of objects */
	obj_list = NULL;
	for(glist = clist->selection; glist != NULL; glist = g_list_next(glist))
	{
		i = (gint)glist->data;
		obj_list = g_list_append(
			obj_list,
			gtk_clist_get_row_data(clist, i)
		);
	}
	if(obj_list == NULL)
	{
#ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_warning(edv_ctx);
#endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"No Images",
			"There are no selected images to convert.",
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		return;
	}

	/* Create a new conversion context */
	ctx = WinConvertConvContextNew();
	if(ctx == NULL)
	{
		g_list_free(obj_list);
		return;
	}

	/* Set the values of the conversion parameters to the
	 * conversion context
	 */
	WinConvertSetParametersToContext(ctx, win->convparm_list);

	/* Start converting */
	WinConvertSetStart(win);

	WinStatusMessage(win, "Converting...", TRUE);

	/* Iterate through each object in the list */
	obj_converted_list = NULL;
	for(i = 0,
		nobjs = g_list_length(obj_list),
		glist = obj_list;
		glist != NULL;
		i++,
		glist = g_list_next(glist)
	)
	{
		if(win->stop_count > 0)
			break;

		obj = OBJ(glist->data);
		if(obj == NULL)
			continue;

		/* Convert this object */
		status = WinConvertIterate(
			ctx, core, win, obj,
			tar_dir,
			recursive, &yes_to_all, &nconverted,
			toplevel
		);
		if(status)
		{
			/* User aborted/canceled? */
			if(status == -4)
				break;
		}
		else
		{
			/* Converted successfully, add this object to the list
			 * of successfully converted objects
			 */
			obj_converted_list = g_list_append(
				obj_converted_list, obj
			);
		}
	}

	/* Print conversion results */
	if(nconverted > 0)
	{
		gchar *s = g_strdup_printf(
			"Converted %i images%s",
			nconverted,
			win->stop_count ? " (interrupted)" : ""
		);
		WinStatusMessage(win, s, TRUE);
		g_free(s);
	}
	else
	{
		WinStatusMessage(
			win,
			"No images converted",
			TRUE
		);
	}

	/* Remove converted? */
	if((obj_converted_list != NULL) && remove_converted)
	{
		for(glist = obj_converted_list;
			glist != NULL;
			glist = g_list_next(glist)
		)
		{
			i = gtk_clist_find_row_from_data(clist, glist->data);
			if(i < 0)
				continue;

			WinSrcListRemove(win, i);
		}
	}

	/* End conversion */
	WinConvertConvContextDelete(ctx);
	g_list_free(obj_converted_list);
	g_list_free(obj_list);
	WinConvertSetEnd(win);

#ifdef HAVE_LIBENDEAVOUR2
	/* Play the convert completed sound */
	edv_play_sound_completed(edv_ctx);
#endif

	/* Update the displayed target preview as needed */
	clist = (GtkCList *)win->tar_clist;
	if(win->preview_tar &&
	   (g_list_length(clist->selection) == 1)
	)
	{
		GList *glist = clist->selection;
		obj_struct *obj = OBJ(gtk_clist_get_row_data(
			clist, (gint)glist->data
		));
		if((obj != NULL) ? !STRISEMPTY(obj->full_path) : FALSE)
		{
			const char *path = obj->full_path;
			win->freeze_count++;
#ifdef HAVE_LIBENDEAVOUR2
			WinPreviewSetImageFile(
				win->tar_preview, path,
				core->edv_ctx
			);
#else
			WinPreviewSetImageFile(
				win->tar_preview, path
			);
#endif
			win->freeze_count--;
		}
	}
}
